"""This module implements basic kinds of jobs for Cp2k runs."""

import logging
import os
import shutil
import subprocess

from monty.os.path import zpath
from monty.shutil import decompress_dir
from pymatgen.io.cp2k.inputs import Cp2kInput, Keyword

from custodian.cp2k.interpreter import Cp2kModder
from custodian.cp2k.utils import cleanup_input, restart
from custodian.custodian import Job

logger = logging.getLogger(__name__)

__author__ = "Nicholas Winner"
__version__ = "1.0"

CP2K_INPUT_FILES = ["cp2k.inp"]
CP2K_OUTPUT_FILES = ["cp2k.out"]


class Cp2kJob(Job):
    """
    A basic cp2k job. Just runs whatever is in the directory. But conceivably
    can be a complex processing of inputs etc. with initialization.
    """

    def __init__(
        self,
        cp2k_cmd,
        input_file="cp2k.inp",
        output_file="cp2k.out",
        stderr_file="std_err.txt",
        suffix="",
        final=True,
        backup=True,
        settings_override=None,
        restart=False,
    ):
        """
        This constructor is necessarily complex due to the need for
        flexibility. For standard kinds of runs, it's often better to use one
        of the static constructors. The defaults are usually fine too.

        Args:
            cp2k_cmd (list): Command to run cp2k as a list of args. For example,
                if you are using mpirun, it can be something like
                ["mpirun", "cp2k.popt"]
            input_file (str): Name of the file to use as input to CP2K
                executable. Defaults to "cp2k.inp"
            output_file (str): Name of file to direct standard out to.
                Defaults to "cp2k.out".
            stderr_file (str): Name of file to direct standard error to.
                Defaults to "std_err.txt".
            suffix (str): A suffix to be appended to the final output. E.g.,
                to rename all CP2K output from say cp2k.out to
                cp2k.out.relax1, provide ".relax1" as the suffix.
            final (bool): Indicating whether this is the final cp2k job in a
                series. Defaults to True.
            backup (bool): Whether to backup the initial input files. If True,
                the input file will be copied with a
                ".orig" appended. Defaults to True.
            settings_override ([actions]): A list of actions. See the Cp2kModder
                in interpreter.py
            restart (bool): Whether to run in restart mode, i.e. this a continuation of
                a previous calculation. Default is False.

        """
        self.cp2k_cmd = cp2k_cmd
        self.input_file = input_file
        self.ci = None
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.final = final
        self.backup = backup
        self.suffix = suffix
        self.settings_override = settings_override if settings_override else []
        self.restart = restart

    def setup(self):
        """
        Performs initial setup for Cp2k in three stages. First, if custodian is running in restart mode, then
        the restart function will copy the restart file to self.input_file, and remove any previous WFN initialization
        if present. Second, any additional user specified settings will be applied. Lastly, a backup of the input
        file will be made for reference.
        """
        decompress_dir(".")

        self.ci = Cp2kInput.from_file(zpath(self.input_file))
        cleanup_input(self.ci)

        if self.restart:
            restart(
                actions=self.settings_override,
                output_file=self.output_file,
                input_file=self.input_file,
                no_actions_needed=True,
            )

        if self.settings_override or self.restart:
            modder = Cp2kModder(filename=self.input_file, actions=[], ci=self.ci)
            modder.apply_actions(self.settings_override)

        if self.backup:
            shutil.copy(self.input_file, f"{self.input_file}.orig")

    def run(self):
        """
        Perform the actual CP2K run.

        Returns:
            (subprocess.Popen) Used for monitoring.
        """
        # TODO: cp2k has bizarre in/out streams. Some errors that should go to std_err are not sent anywhere...
        cmd = list(self.cp2k_cmd)
        cmd.extend(["-i", self.input_file])
        cmd_str = " ".join(cmd)
        logger.info(f"Running {cmd_str}")
        with open(self.output_file, "w") as f_std, open(self.stderr_file, "w", buffering=1) as f_err:
            # use line buffering for stderr
            return subprocess.Popen(cmd, stdout=f_std, stderr=f_err, shell=False)

    # TODO double jobs, file manipulations, etc. should be done in atomate in the future
    # and custodian should only run the job itself
    def postprocess(self):
        """Postprocessing includes renaming and gzipping where necessary."""
        files = os.listdir(".")
        if os.path.exists(self.output_file) and self.suffix != "":
            os.mkdir(f"run{self.suffix}")
            for file in files:
                if "json" in file:
                    continue
                if not os.path.isdir(file):
                    if self.final:
                        shutil.move(file, f"run{self.suffix}/{file}")
                    else:
                        shutil.copy(file, f"run{self.suffix}/{file}")

        # Remove continuation so if a subsequent job is run in
        # the same directory, will not restart this job.
        if os.path.exists("continue.json"):
            os.remove("continue.json")

    def terminate(self):
        """Terminate cp2k."""
        for cmd in self.cp2k_cmd:
            if "cp2k" in cmd:
                try:
                    os.system(f"killall {cmd}")
                except Exception:
                    pass

    @classmethod
    def gga_static_to_hybrid(
        cls,
        cp2k_cmd,
        input_file="cp2k.inp",
        output_file="cp2k.out",
        stderr_file="std_err.txt",
        backup=True,
        settings_override_gga=None,
        settings_override_hybrid=None,
    ):
        """
        A bare GGA to hybrid calculation. Removes all unnecessary features
        from the GGA run, and making it only a ENERGY/ENERGY_FORCE
        depending on the hybrid run.
        """
        job1_settings_override = [
            {
                "dict": input_file,
                "action": {
                    "_unset": {"FORCE_EVAL": {"DFT": "XC"}},
                    "_set": {"GLOBAL": {"PROJECT_NAME": "GGA", "RUN_TYPE": "ENERGY_FORCE"}},
                },
            },
            {
                "dict": input_file,
                "action": {"_set": {"FORCE_EVAL": {"DFT": {"XC": {"XC_FUNCTIONAL": {"PBE": {}}}}}}},
            },
        ]

        job1 = Cp2kJob(
            cp2k_cmd,
            input_file=input_file,
            output_file=output_file,
            backup=backup,
            stderr_file=stderr_file,
            final=False,
            suffix="1",
            settings_override=job1_settings_override,
        )

        ci = Cp2kInput.from_file(zpath(input_file))
        run_type = ci["global"].get("run_type", Keyword("RUN_TYPE", "ENERGY_FORCE")).values[0]
        if run_type in {"ENERGY", "WAVEFUNCTION_OPTIMIZATION", "WFN_OPT", "ENERGY_FORCE"}:  # no need for double job
            return [job1]

        job2_settings_override = [
            {
                "dict": input_file,
                "action": {
                    "_set": {
                        "FORCE_EVAL": {
                            "DFT": {
                                "XC": {
                                    "HF": {
                                        "SCREENING": {
                                            "SCREEN_ON_INITIAL_P": True,
                                            "SCREEN_P_FORCES": True,
                                        }
                                    }
                                },
                                "WFN_RESTART_FILE_NAME": "GGA-RESTART.wfn",
                            }
                        },
                        "GLOBAL": {"RUN_TYPE": run_type},
                    },
                },
            }
        ]

        job2 = Cp2kJob(
            cp2k_cmd,
            input_file=input_file,
            output_file=output_file,
            backup=backup,
            stderr_file=stderr_file,
            final=True,
            suffix="2",
            restart=False,
            settings_override=job2_settings_override,
        )
        return [job1, job2]

    @classmethod
    def double_job(
        cls, cp2k_cmd, input_file="cp2k.inp", output_file="cp2k.out", stderr_file="std_err.txt", backup=True
    ):
        """
        This creates a sequence of two jobs. The first of which is an "initialization" of the
        wfn. Using this, the "restart" function can be exploited to determine if a diagonalization
        job can/would benefit from switching to OT scheme. If not, then the second job remains a
        diagonalization job, and there is minimal overhead from restarting.
        """
        job1 = Cp2kJob(
            cp2k_cmd,
            input_file=input_file,
            output_file=output_file,
            backup=backup,
            stderr_file=stderr_file,
            final=False,
            suffix="1",
            settings_override={},
        )
        ci = Cp2kInput.from_file(zpath(input_file))
        run_type = ci["global"].get("run_type", Keyword("RUN_TYPE", "ENERGY_FORCE")).values[0]
        if run_type not in {"ENERGY", "WAVEFUNCTION_OPTIMIZATION", "WFN_OPT"}:
            job1.settings_override = [
                {"dict": input_file, "action": {"_set": {"GLOBAL": {"RUN_TYPE": "ENERGY_FORCE"}}}}
            ]

        job2 = Cp2kJob(
            cp2k_cmd,
            input_file=input_file,
            output_file=output_file,
            backup=backup,
            stderr_file=stderr_file,
            final=True,
            suffix="2",
            restart=True,
        )
        job2.settings_override = [{"dict": input_file, "action": {"_set": {"GLOBAL": {"RUN_TYPE": run_type}}}}]

        return [job1, job2]

    @classmethod
    def pre_screen_hybrid(
        cls, cp2k_cmd, input_file="cp2k.inp", output_file="cp2k.out", stderr_file="std_err.txt", backup=True
    ):
        """
        Build a job where the first job is an unscreened hybrid static calculation, then the second one
        uses the wfn from the first job as a restart to do a screened calculation.
        """
        job1_settings_override = [
            {
                "dict": input_file,
                "action": {
                    "_set": {
                        "FORCE_EVAL": {
                            "DFT": {
                                "XC": {
                                    "HF": {
                                        "SCREENING": {
                                            "SCREEN_ON_INITIAL_P": False,
                                            "SCREEN_P_FORCES": False,
                                        }
                                    }
                                }
                            }
                        },
                        "GLOBAL": {"RUN_TYPE": "ENERGY_FORCE"},
                    }
                },
            }
        ]

        job1 = Cp2kJob(
            cp2k_cmd,
            input_file=input_file,
            output_file=output_file,
            backup=backup,
            stderr_file=stderr_file,
            final=False,
            suffix="1",
            settings_override=job1_settings_override,
        )

        ci = Cp2kInput.from_file(zpath(input_file))
        r = ci["global"].get("run_type", Keyword("RUN_TYPE", "ENERGY_FORCE")).values[0]
        if r in {"ENERGY", "WAVEFUNCTION_OPTIMIZATION", "WFN_OPT", "ENERGY_FORCE"}:  # no need for double job
            return [job1]

        job2_settings_override = [
            {
                "dict": input_file,
                "action": {
                    "_set": {
                        "FORCE_EVAL": {
                            "DFT": {
                                "XC": {
                                    "HF": {
                                        "SCREENING": {
                                            "SCREEN_ON_INITIAL_P": True,
                                            "SCREEN_P_FORCES": True,
                                        }
                                    }
                                },
                                "WFN_RESTART_FILE_NAME": "UNSCREENED_HYBRID-RESTART.wfn",
                            }
                        },
                        "GLOBAL": {"RUN_TYPE": r},
                    },
                },
            }
        ]

        job2 = Cp2kJob(
            cp2k_cmd,
            input_file=input_file,
            output_file=output_file,
            backup=backup,
            stderr_file=stderr_file,
            final=True,
            suffix="2",
            restart=False,
            settings_override=job2_settings_override,
        )

        return [job1, job2]



"""
The ansible package provides modules that provides a mongo-like syntax for
making modifications to dicts, objects and files. The mongo-like syntax
itself is a dict.

The main use of this package is to allow changes to objects or files to be
stored in a json file or MongoDB database, i.e., a form of version control
or tracked changes (though without undo capability unless the input is
stored at each step).
"""

from .actions import DictActions, FileActions
from .interpreter import Modder



"""This is a script to converge the geometry of a system."""

import logging

from pymatgen.io.vasp.outputs import Vasprun

from custodian.custodian import Custodian
from custodian.vasp.handlers import (
    MeshSymmetryErrorHandler,
    NonConvergingErrorHandler,
    PotimErrorHandler,
    UnconvergedErrorHandler,
    VaspErrorHandler,
)
from custodian.vasp.jobs import VaspJob

FORMAT = "%(asctime)s %(message)s"
logging.basicConfig(format=FORMAT, level=logging.INFO, filename="run.log")


def get_runs(args):
    """Get the runs."""
    vasp_command = args.command.split()
    converged = False
    job_number = 0

    while (not converged) and (job_number < args.max_relax):
        suffix = f".relax{job_number + 1}"

        if job_number == 0:
            backup = True
            # assume the initial guess is poor,
            # start with conjugate gradients
            settings = [{"dict": "INCAR", "action": {"_set": {"IBRION": 2}}}]

        else:
            backup = False
            v = Vasprun("vasprun.xml")

            if len(v.ionic_steps) == 1:
                converged = True

            if job_number < 2 and not converged:
                settings = [
                    {"dict": "INCAR", "action": {"_set": {"ISTART": 1}}},
                    {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}},
                ]

            # switch to RMM-DIIS once we are near the
            # local minimum (assumed after 2 runs of CG)
            else:
                settings = [
                    {"dict": "INCAR", "action": {"_set": {"ISTART": 1, "IBRION": 1}}},
                    {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}},
                ]

        job_number += 1
        yield VaspJob(
            vasp_command,
            final=converged,
            backup=backup,
            suffix=suffix,
            settings_override=settings,
        )


def do_run(args):
    """Perform the run."""
    handlers = [
        VaspErrorHandler(),
        MeshSymmetryErrorHandler(),
        UnconvergedErrorHandler(),
        NonConvergingErrorHandler(),
        PotimErrorHandler(),
    ]
    c = Custodian(handlers, get_runs(args), max_errors=10, gzipped_output=args.gzip)
    c.run()
    logging.info("Geometry optimization complete")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="""
    converge_geometry performs a geometry optimization. What this script will do
    is run a particular VASP relaxation repeatedly until the geometry
    is converged within the first ionic step. This is a common practice for
    converging molecular geometries in VASP, especially in situations where
    the geometry needs to be precise: such as frequency calculations.
    """,
        epilog="Author: Stephen Dacek",
    )

    parser.add_argument(
        "-c",
        "--command",
        dest="command",
        nargs="?",
        default="pvasp",
        type=str,
        help="VASP command. Defaults to pvasp. If you are using mpirun, set this to something like 'mpirun pvasp'.",
    )

    parser.add_argument(
        "-z",
        "--gzip",
        dest="gzip",
        action="store_true",
        help="Add this option to gzip the final output. Do not gzip if you "
        "are going to perform an additional static run.",
    )

    parser.add_argument(
        "-mr",
        "--max_relaxtions",
        dest="max_relax",
        default=10,
        type=int,
        help="Maximum number of relaxations to allow",
    )

    args = parser.parse_args()
    do_run(args)



"""This module implements jobs for Lobster runs."""

import logging
import os
import shutil
import subprocess

from monty.io import zopen
from monty.shutil import compress_file

from custodian.custodian import Job

__author__ = "Janine George, Guido Petretto,Aakash Naik"
__copyright__ = "Copyright 2020, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Janine George"
__email__ = "janine.george@uclouvain.be"
__date__ = "April 27, 2020"

LOBSTERINPUT_FILES = ["lobsterin"]
LOBSTEROUTPUT_FILES = [
    "lobsterout",
    "CHARGE.lobster",
    "COHPCAR.lobster",
    "COOPCAR.lobster",
    "DOSCAR.lobster",
    "GROSSPOP.lobster",
    "ICOHPLIST.lobster",
    "ICOOPLIST.lobster",
    "lobster.out",
    "projectionData.lobster",
    "MadelungEnergies.lobster",
    "SitePotentials.lobster",
    "bandOverlaps.lobster",
    "ICOBILIST.lobster",
    "COBICAR.lobster",
    "DOSCAR.LSO.lobster",
]
FW_FILES = ["custodian.json", "FW.json", "FW_submit.script"]

logger = logging.getLogger(__name__)


class LobsterJob(Job):
    """Runs the Lobster Job."""

    def __init__(
        self,
        lobster_cmd: str,
        output_file: str = "lobsterout",
        stderr_file: str = "std_err_lobster.txt",
        gzipped: bool = True,
        add_files_to_gzip=(),
        backup: bool = True,
    ):
        """

        Args:
            lobster_cmd: command to run lobster
            output_file: usually lobsterout
            stderr_file: standard output
            gzipped: if True, Lobster files and add_files_to_gzip will be gzipped
            add_files_to_gzip: list of files that should be gzipped
            backup: if True, lobsterin will be copied to lobsterin.orig.
        """
        self.lobster_cmd = lobster_cmd
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.gzipped = gzipped
        self.add_files_to_gzip = add_files_to_gzip
        self.backup = backup

    def setup(self):
        """Will backup lobster input files."""
        if self.backup:
            for file in LOBSTERINPUT_FILES:
                shutil.copy(file, f"{file}.orig")

    def run(self):
        """Runs the job."""
        cmd = self.lobster_cmd

        logger.info(f"Running {' '.join(cmd)}")

        with zopen(self.output_file, "w") as f_std, zopen(self.stderr_file, "w", buffering=1) as f_err:
            # use line buffering for stderr
            return subprocess.Popen(cmd, stdout=f_std, stderr=f_err)  # pylint: disable=R1732

    def postprocess(self):
        """Will gzip relevant files (won't gzip custodian.json and other output files from the cluster)."""
        if self.gzipped:
            for file in LOBSTEROUTPUT_FILES:
                if os.path.exists(file):
                    compress_file(file, compression="gz")
            for file in LOBSTERINPUT_FILES:
                if os.path.exists(file):
                    compress_file(file, compression="gz")
            if self.backup and os.path.exists("lobsterin.orig"):
                compress_file("lobsterin.orig", compression="gz")
            for file in FW_FILES:
                if os.path.exists(file):
                    compress_file(file, compression="gz")
            for file in self.add_files_to_gzip:
                compress_file(file, compression="gz")



"""This module holds different utility functions. Mainly used by handlers."""

import itertools
import os
from collections import deque

from pymatgen.io.cp2k.inputs import Cp2kInput
from pymatgen.io.cp2k.outputs import Cp2kOutput


def restart(actions, output_file, input_file, no_actions_needed=False):
    """
    Helper function. To discard old restart if convergence is already good, and copy
    the restart file to the input file. Restart also supports switching back and forth
    between OT and diagonalization as needed based on convergence behavior. If OT is not
    being used and a band gap exists, then OT will be activated.

    Args:
        actions (list): list of actions that the handler is going to return to custodian. If
            no actions are present, then non are added by this function
        output_file (str): the cp2k output file name.
        input_file (str): the cp2k input file name.
        no_actions_needed (bool): if no actions are needed, then this should be set to True.
    """
    if actions or no_actions_needed:
        out = Cp2kOutput(output_file)
        ci = Cp2kInput.from_file(input_file)
        restart_file = out.filenames.get("restart")
        restart_file = restart_file[-1] if restart_file else None
        wfn_restart = ci["force_eval"]["dft"].get("wfn_restart_file_name") if ci.check("force_eval/dft") else None

        # If convergence is already pretty good, or we have moved to a new ionic step,
        # discard the old WFN
        if wfn_restart:
            conv = get_conv(output_file)
            if (conv and conv[-1] <= 1e-5) or restart_file:
                actions.append(
                    {"dict": input_file, "action": {"_unset": {"FORCE_EVAL": {"DFT": "WFN_RESTART_FILE_NAME"}}}}
                )

        # If issues arose after some ionic steps and corrections are possible
        # then switch the restart file to the input file.
        if restart_file:
            actions.insert(
                0,
                {
                    "file": os.path.abspath(restart_file),
                    "action": {"_file_copy": {"dest": os.path.abspath(input_file)}},
                },
            )


# TODO Not sure I like this solution
def cleanup_input(ci):
    """
    Intention is to use this to remove problematic parts of the input file.

        (1) The "POTENTIAL" section within KIND cannot be empty, but the number
            sequences used inside do not play nice with the input parser

    """
    if not hasattr(ci, "subsections") or not ci.subsections:
        return
    if any(k.upper() == "POTENTIAL" for k in ci.subsections):
        ci.subsections.pop("POTENTIAL")
    for v in ci.subsections.values():
        cleanup_input(v)


def activate_ot(actions, ci):
    """
    Activate OT scheme.

    actions (list):
        list of actions that are being applied. Will be modified in-place

    ci (Cp2kInput):
        Cp2kInput object, used to coordinate settings
    """
    eps_scf = ci["force_eval"]["dft"]["scf"]["eps_scf"]

    ot_actions = [
        {
            "dict": "cp2k.inp",
            "action": [
                (
                    "_unset",
                    {"FORCE_EVAL": {"DFT": "SCF"}},
                )
            ],
        },
        {
            "dict": "cp2k.inp",
            "action": [
                (
                    "_set",
                    {
                        "FORCE_EVAL": {
                            "DFT": {
                                "SCF": {
                                    "MAX_SCF": 20,
                                    "OT": {
                                        "ENERGY_GAP": 0.01,
                                        "ALGORITHM": "STRICT",
                                        "PRECONDITIONER": "FULL_ALL",
                                        "MINIMIZER": "DIIS",
                                        "LINESEARCH": "2PNT",
                                    },
                                    "OUTER_SCF": {"MAX_SCF": 20, "EPS_SCF": eps_scf},
                                }
                            }
                        }
                    },
                )
            ],
        },
    ]
    actions.extend(ot_actions)


def activate_diag(actions):
    """
    Activate diagonalization.

    actions (list):
        list of actions that are being applied. Will be modified in-place
    """
    diag_actions = [
        {"dict": "cp2k.inp", "action": ("_unset", {"FORCE_EVAL": {"DFT": {"SCF": "OT"}}})},
        {"dict": "cp2k.inp", "action": ("_unset", {"FORCE_EVAL": {"DFT": {"SCF": "OUTER_SCF"}}})},
        {
            "dict": "cp2k.inp",
            "action": (
                "_set",
                {
                    "FORCE_EVAL": {
                        "DFT": {
                            "SCF": {
                                "MAX_SCF": 200,
                                "ADDED_MOS": 100,  # TODO needs to be dynamic value
                                "MAX_DIIS": 15,
                                "DIAGONALIZATION": {},
                                "MIXING": {"ALPHA": 0.05},
                                "SMEAR": {"ELEC_TEMP": 300, "METHOD": "FERMI_DIRAC"},
                            }
                        }
                    }
                },
            ),
        },
    ]
    actions.extend(diag_actions)


def can_use_ot(output, ci, minimum_band_gap=0.1):
    """
    Check whether OT can be used:
        OT should not already be activated
        The output should show that the system has a band gap that is greater than minimum_band_gap.

    Args:
        output (Cp2kOutput): cp2k output object for determining band gap
        ci (Cp2kInput): cp2k input object for determining if OT is already active
        minimum_band_gap (float): the minimum band gap for OT
    """
    output.parse_dos()
    if (
        not ci.check("FORCE_EVAL/DFT/SCF/OT")
        and not ci.check("FORCE_EVAL/DFT/KPOINTS")
        and output.band_gap
        and output.band_gap > minimum_band_gap
    ):
        return True
    return False


def tail(filename, n=10):
    """Returns the last n lines of a file as a list (including empty lines)."""
    with open(filename) as f:
        t = deque(f, n)
        if t:
            return t
        return [""] * n


def get_conv(outfile):
    """
    Helper function to get the convergence info from SCF loops.

    Args:
        outfile (str): output file to parse
    Returns:
        returns convergence info (change in energy between SCF steps) as a
        single list (flattened across outer scf loops).
    """
    out = Cp2kOutput(outfile, auto_load=False, verbose=False)
    out.parse_scf_opt()
    return list(itertools.chain.from_iterable(out.data["convergence"]))



import os
import random
import subprocess
import unittest
from glob import glob

import pytest
from ruamel.yaml import YAML

from custodian.custodian import (
    Custodian,
    ErrorHandler,
    Job,
    MaxCorrectionsError,
    MaxCorrectionsPerHandlerError,
    MaxCorrectionsPerJobError,
    NonRecoverableError,
    ReturnCodeError,
    ValidationError,
    Validator,
)


class ExitCodeJob(Job):
    def __init__(self, exitcode=0):
        self.exitcode = exitcode

    def setup(self):
        pass

    def run(self):
        return subprocess.Popen(f"exit {self.exitcode}", shell=True)

    def postprocess(self):
        pass


class ExampleJob(Job):
    def __init__(self, jobid, params=None):
        if params is None:
            params = {"initial": 0, "total": 0}
        self.jobid = jobid
        self.params = params

    def setup(self):
        self.params["initial"] = 0
        self.params["total"] = 0

    def run(self):
        sequence = [random.uniform(0, 1) for i in range(100)]
        self.params["total"] = self.params["initial"] + sum(sequence)

    def postprocess(self):
        pass

    @property
    def name(self):
        return f"ExampleJob{self.jobid}"


class ExampleHandler(ErrorHandler):
    def __init__(self, params):
        self.params = params

    def check(self):
        return self.params["total"] < 50

    def correct(self):
        self.params["initial"] += 1
        return {"errors": "total < 50", "actions": "increment by 1"}


class ExampleHandler1b(ExampleHandler):
    """
    This handler always can apply a correction, but will only apply it twice before raising.
    """

    max_num_corrections = 2  # type: ignore
    raise_on_max = True


class ExampleHandler1c(ExampleHandler):
    """
    This handler always can apply a correction, but will only apply it twice and then not anymore.
    """

    max_num_corrections = 2  # type: ignore
    raise_on_max = False


class ExampleHandler2(ErrorHandler):
    """
    This handler always result in an error.
    """

    def __init__(self, params):
        self.params = params
        self.has_error = False

    def check(self):
        return True

    def correct(self):
        self.has_error = True
        return {"errors": "Unrecoverable error", "actions": None}


class ExampleHandler2b(ExampleHandler2):
    """
    This handler always result in an error. No runtime error though
    """

    raises_runtime_error = False

    def correct(self):
        self.has_error = True
        return {"errors": "Unrecoverable error", "actions": []}


class ExampleValidator1(Validator):
    def __init__(self):
        pass

    def check(self):
        return False


class ExampleValidator2(Validator):
    def __init__(self):
        pass

    def check(self):
        return True


class CustodianTest(unittest.TestCase):
    def setUp(self):
        self.cwd = os.getcwd()
        os.chdir(os.path.abspath(os.path.dirname(__file__)))

    def test_exitcode_error(self):
        c = Custodian([], [ExitCodeJob(0)])
        c.run()
        c = Custodian([], [ExitCodeJob(1)])
        with pytest.raises(
            ReturnCodeError,
        ):
            c.run()
        assert c.run_log[-1]["nonzero_return_code"]
        c = Custodian([], [ExitCodeJob(1)], terminate_on_nonzero_returncode=False)
        c.run()

    def test_run(self):
        n_jobs = 100
        params = {"initial": 0, "total": 0}
        c = Custodian(
            [ExampleHandler(params)],
            [ExampleJob(i, params) for i in range(n_jobs)],
            max_errors=n_jobs,
        )
        output = c.run()
        assert len(output) == n_jobs
        ExampleHandler(params).as_dict()

    def test_run_interrupted(self):
        n_jobs = 100
        params = {"initial": 0, "total": 0}
        c = Custodian(
            [ExampleHandler(params)],
            [ExampleJob(i, params) for i in range(n_jobs)],
            max_errors=n_jobs,
        )

        assert c.run_interrupted() == n_jobs
        assert c.run_interrupted() == n_jobs

        total_done = 1
        while total_done < n_jobs:
            c.jobs[n_jobs - 1].run()
            if params["total"] > 50:
                assert c.run_interrupted() == n_jobs - total_done
                total_done += 1

    def test_unrecoverable(self):
        n_jobs = 100
        params = {"initial": 0, "total": 0}
        handler = ExampleHandler2(params)
        c = Custodian([handler], [ExampleJob(i, params) for i in range(n_jobs)], max_errors=n_jobs)
        with pytest.raises(NonRecoverableError):
            c.run()
        assert handler.has_error
        handler = ExampleHandler2b(params)
        c = Custodian([handler], [ExampleJob(i, params) for i in range(n_jobs)], max_errors=n_jobs)
        c.run()
        assert handler.has_error
        assert c.run_log[-1]["handler"] == handler

    def test_max_errors(self):
        n_jobs = 100
        params = {"initial": 0, "total": 0}
        handler = ExampleHandler(params)
        c = Custodian(
            [handler],
            [ExampleJob(i, params) for i in range(n_jobs)],
            max_errors=1,
            max_errors_per_job=10,
        )
        with pytest.raises(MaxCorrectionsError):
            c.run()
        assert c.run_log[-1]["max_errors"]

    def test_max_errors_per_job(self):
        n_jobs = 100
        params = {"initial": 0, "total": 0}
        handler = ExampleHandler(params)
        c = Custodian(
            [handler],
            [ExampleJob(i, params) for i in range(n_jobs)],
            max_errors=n_jobs,
            max_errors_per_job=1,
        )
        with pytest.raises(MaxCorrectionsPerJobError):
            c.run()
        assert c.run_log[-1]["max_errors_per_job"]

    def test_max_errors_per_handler_raise(self):
        n_jobs = 100
        params = {"initial": 0, "total": 0}
        handler = ExampleHandler1b(params)
        c = Custodian(
            [handler],
            [ExampleJob(i, params) for i in range(n_jobs)],
            max_errors=n_jobs * 10,
            max_errors_per_job=1000,
        )
        with pytest.raises(MaxCorrectionsPerHandlerError):
            c.run()
        assert handler.n_applied_corrections == 2
        assert len(c.run_log[-1]["corrections"]) == 2
        assert c.run_log[-1]["max_errors_per_handler"]
        assert c.run_log[-1]["handler"] == handler

    def test_max_errors_per_handler_warning(self):
        n_jobs = 100
        params = {"initial": 0, "total": 0}
        c = Custodian(
            [ExampleHandler1c(params)],
            [ExampleJob(i, params) for i in range(n_jobs)],
            max_errors=n_jobs * 10,
            max_errors_per_job=1000,
        )
        c.run()
        assert all(len(r["corrections"]) <= 2 for r in c.run_log)

    def test_validators(self):
        n_jobs = 100
        params = {"initial": 0, "total": 0}
        c = Custodian(
            [ExampleHandler(params)],
            [ExampleJob(i, params) for i in range(n_jobs)],
            [ExampleValidator1()],
            max_errors=n_jobs,
        )
        output = c.run()
        assert len(output) == n_jobs

        n_jobs = 100
        params = {"initial": 0, "total": 0}
        v = ExampleValidator2()
        c = Custodian(
            [ExampleHandler(params)],
            [ExampleJob(i, params) for i in range(n_jobs)],
            [v],
            max_errors=n_jobs,
        )
        with pytest.raises(ValidationError):
            c.run()
        assert c.run_log[-1]["validator"] == v

    def test_from_spec(self):
        spec = """jobs:
- jb: custodian.vasp.jobs.VaspJob
  params:
    final: False
    suffix: .relax1
- jb: custodian.vasp.jobs.VaspJob
  params:
    final: True
    suffix: .relax2
    settings_override: {"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}}
jobs_common_params:
  $vasp_cmd: ["mpirun", "-machinefile", "$PBS_NODEFILE", "-np", "24", "/opt/vasp/5.4.1/bin/vasp"]
handlers:
- hdlr: custodian.vasp.handlers.VaspErrorHandler
- hdlr: custodian.vasp.handlers.AliasingErrorHandler
- hdlr: custodian.vasp.handlers.MeshSymmetryErrorHandler
validators:
- vldr: custodian.vasp.validators.VasprunXMLValidator
custodian_params:
  $scratch_dir: $TMPDIR"""

        os.environ["TMPDIR"] = "/tmp/random"
        os.environ["PBS_NODEFILE"] = "whatever"
        dct = YAML(typ="rt").load(spec)
        c = Custodian.from_spec(dct)
        assert c.jobs[0].vasp_cmd[2] == "whatever"
        assert c.scratch_dir == "/tmp/random"
        assert len(c.jobs) == 2
        assert len(c.handlers) == 3
        assert len(c.validators) == 1

    def tearDown(self):
        for file in glob("custodian.*.tar.gz"):
            os.remove(file)
        try:
            os.remove("custodian.json")
        except OSError:
            pass  # Ignore if file cannot be found.
        os.chdir(self.cwd)


# class TestCustodianCheckpoint:
#     def test_checkpoint_loading(self):
#         self.cwd = os.getcwd()
#         pth = f"{TEST_FILES}/checkpointing"
#         # os.chdir()
#         shutil.copy(os.path.join(pth, "backup.tar.gz"), "custodian.chk.3.tar.gz")
#         n_jobs = 5
#         params = {"initial": 0, "total": 0}
#         c = Custodian(
#             [ExampleHandler(params)],
#             [ExampleJob(i, params) for i in range(n_jobs)],
#             [ExampleValidator1()],
#             max_errors=100,
#             checkpoint=True,
#         )
#         assert len(c.run_log) == 3
#         assert len(c.run()) == 5
#         os.remove("custodian.json")
#         os.chdir(self.cwd)



"""Utility function and classes."""

import functools
import logging
import os
import tarfile
from glob import glob


def backup(filenames, prefix="error"):
    """
    Backup files to a tar.gz file. Used, for example, in backing up the
    files of an errored run before performing corrections.

    Args:
        filenames ([str]): List of files to backup. Supports wildcards, e.g.,
            *.*.
        prefix (str): prefix to the files. Defaults to error, which means a
            series of error.1.tar.gz, error.2.tar.gz, ... will be generated.
    """
    num = max([0] + [int(file.split(".")[1]) for file in glob(f"{prefix}.*.tar.gz")])
    filename = f"{prefix}.{num + 1}.tar.gz"
    logging.info(f"Backing up run to {filename}.")
    with tarfile.open(filename, "w:gz") as tar:
        for fname in filenames:
            for file in glob(fname):
                tar.add(file)


def get_execution_host_info():
    """
    Tries to return a tuple describing the execution host.
    Doesn't work for all queueing systems.

    Returns:
        (HOSTNAME, CLUSTER_NAME)
    """
    host = os.environ.get("HOSTNAME", None)
    cluster = os.environ.get("SGE_O_HOST", None)
    if host is None:
        try:
            import socket

            host = host or socket.gethostname()
        except Exception:
            pass
    return host or "unknown", cluster or "unknown"


class tracked_lru_cache:
    """
    Decorator wrapping the functools.lru_cache adding a tracking of the
    functions that have been wrapped.

    Exposes a method to clear the cache of all the wrapped functions.

    Used to cache the parsed outputs in handlers/validators, to avoid
    multiple parsing of the same file.
    Allows Custodian to clear the cache after all the checks have been performed.
    """

    cached_functions: set = set()

    def __init__(self, func):
        """
        Args:
            func: function to be decorated.
        """
        self.func = functools.lru_cache(func)
        functools.update_wrapper(self, func)

        # expose standard lru_cache functions
        self.cache_info = self.func.cache_info
        self.cache_clear = self.func.cache_clear

    def __call__(self, *args, **kwargs):
        """Call the decorated function."""
        result = self.func(*args, **kwargs)
        self.cached_functions.add(self.func)
        return result

    @classmethod
    def tracked_cache_clear(cls):
        """Clear the cache of all the decorated functions."""
        while cls.cached_functions:
            f = cls.cached_functions.pop()
            f.cache_clear()



"""
This is a master vasp running script to perform various combinations of VASP
runs.
"""

import logging
import sys

from pymatgen.io.vasp.inputs import Incar, Kpoints, VaspInput
from ruamel.yaml import YAML

from custodian.custodian import Custodian
from custodian.vasp.jobs import VaspJob


def load_class(mod, name):
    """Load the class from mod and name specification."""
    toks = name.split("?")
    params = {}
    if len(toks) == 2:
        for p in toks[-1].split(","):
            ptoks = p.split("=")
            params[ptoks[0]] = YAML(typ="rt").load(ptoks[1])
    elif len(toks) > 2:
        print("Bad handler specification")
        sys.exit(-1)
    mod = __import__(mod, globals(), locals(), [toks[0]], 0)
    return getattr(mod, toks[0])(**params)


def get_jobs(args):
    """Returns a generator of jobs. Allows of "infinite" jobs."""
    vasp_command = args.command.split()
    # save initial INCAR for rampU runs
    n_ramp_u = args.jobs.count("rampU")
    ramps = 0
    if n_ramp_u:
        incar = Incar.from_file("INCAR")
        ldauu = incar["LDAUU"]
        ldauj = incar["LDAUJ"]

    n_jobs = len(args.jobs)
    post_settings = []  # append to this list to have settings applied on next job
    for i, job in enumerate(args.jobs):
        final = i == n_jobs - 1
        suffix = "." + job if any(c.isdigit() for c in job) else f".{job}{i + 1}"
        settings = post_settings
        post_settings = []
        backup = i == 0
        copy_magmom = False
        vinput = VaspInput.from_directory(".")
        if i > 0:
            settings.append({"file": "CONTCAR", "action": {"_file_copy": {"dest": "POSCAR"}}})

        job_type = job.lower()
        auto_npar = True

        if args.no_auto_npar:
            auto_npar = False

        if job_type.startswith("static_derived"):
            from pymatgen.io.vasp.sets import MPStaticSet

            vis = MPStaticSet.from_prev_calc(
                ".",
                user_incar_settings={"LWAVE": True, "EDIFF": 1e-6},
                ediff_per_atom=False,
            )
            settings.extend(
                [
                    {"dict": "INCAR", "action": {"_set": dict(vis.incar)}},
                    {"dict": "KPOINTS", "action": {"_set": vis.kpoints.as_dict()}},
                ]
            )

        if job_type.startswith("static_dielectric_derived"):
            from pymatgen.io.vasp.sets import MPStaticDielectricDFPTVaspInputSet, MPStaticSet

            # vis = MPStaticSet.from_prev_calc(
            #     ".",
            #     user_incar_settings={
            #         "EDIFF": 1e-6,
            #         "IBRION": 8,
            #         "LEPSILON": True,
            #         "LREAL": False,
            #         "LPEAD": True,
            #         "ISMEAR": 0,
            #         "SIGMA": 0.01,
            #     },
            #     ediff_per_atom=False,
            # )
            vis = MPStaticDielectricDFPTVaspInputSet()
            incar = vis.get_incar(vinput["POSCAR"].structure)
            unset = {}
            for k in ["NPAR", "KPOINT_BSE", "LAECHG", "LCHARG", "LVHAR", "NSW"]:
                incar.pop(k, None)
                if k in vinput["INCAR"]:
                    unset[k] = 1
            kpoints = vis.get_kpoints(vinput["POSCAR"].structure)
            settings.extend(
                [
                    {"dict": "INCAR", "action": {"_set": dict(incar), "_unset": unset}},
                    {"dict": "KPOINTS", "action": {"_set": kpoints.as_dict()}},
                ]
            )
            auto_npar = False
        elif job_type.startswith("static") and vinput["KPOINTS"]:
            m = [i * args.static_kpoint for i in vinput["KPOINTS"].kpts[0]]
            settings.extend(
                [
                    {"dict": "INCAR", "action": {"_set": {"NSW": 0}}},
                    {"dict": "KPOINTS", "action": {"_set": {"kpoints": [m]}}},
                ]
            )

        elif job_type.startswith("nonscf_derived"):
            from pymatgen.io.vasp.sets import MPNonSCFSet

            vis = MPNonSCFSet.from_prev_calc(".", copy_chgcar=False, user_incar_settings={"LWAVE": True})
            settings.extend(
                [
                    {"dict": "INCAR", "action": {"_set": dict(vis.incar)}},
                    {"dict": "KPOINTS", "action": {"_set": vis.kpoints.as_dict()}},
                ]
            )

        elif job_type.startswith("optics_derived"):
            from pymatgen.io.vasp.sets import MPNonSCFSet

            vis = MPNonSCFSet.from_prev_calc(
                ".",
                optics=True,
                copy_chgcar=False,
                nedos=2001,
                mode="uniform",
                nbands_factor=5,
                user_incar_settings={
                    "LWAVE": True,
                    "ALGO": "Exact",
                    "SIGMA": 0.01,
                    "EDIFF": 1e-6,
                },
                ediff_per_atom=False,
            )
            settings.extend(
                [
                    {"dict": "INCAR", "action": {"_set": dict(vis.incar)}},
                    {"dict": "KPOINTS", "action": {"_set": vis.kpoints.as_dict()}},
                ]
            )

        elif job_type.startswith("rampu"):
            f = ramps / (n_ramp_u - 1)
            settings.append(
                {
                    "dict": "INCAR",
                    "action": {
                        "_set": {
                            "LDAUJ": [j * f for j in ldauj],
                            "LDAUU": [u * f for u in ldauu],
                        }
                    },
                }
            )
            copy_magmom = True
            ramps += 1
        elif job_type.startswith(("quick_relax", "quickrelax")):
            kpoints = vinput["KPOINTS"]
            incar = vinput["INCAR"]
            structure = vinput["POSCAR"].structure
            if "ISMEAR" in incar:
                post_settings.append({"dict": "INCAR", "action": {"_set": {"ISMEAR": incar["ISMEAR"]}}})
            else:
                post_settings.append({"dict": "INCAR", "action": {"_unset": {"ISMEAR": 1}}})
            post_settings.append({"dict": "KPOINTS", "action": {"_set": kpoints.as_dict()}})
            # lattice vectors with length < 9 will get >1 KPOINT
            low_kpoints = Kpoints.gamma_automatic([max(int(18 / length), 1) for length in structure.lattice.abc])
            settings.extend(
                [
                    {"dict": "INCAR", "action": {"_set": {"ISMEAR": 0}}},
                    {"dict": "KPOINTS", "action": {"_set": low_kpoints.as_dict()}},
                ]
            )

            # let vasp determine encut (will be lower than
            # needed for compatibility with other runs)
            if "ENCUT" in incar:
                post_settings.append({"dict": "INCAR", "action": {"_set": {"ENCUT": incar["ENCUT"]}}})
                settings.append({"dict": "INCAR", "action": {"_unset": {"ENCUT": 1}}})

        elif job_type.startswith("relax"):
            pass
        elif job_type.startswith("full_relax"):
            yield from VaspJob.full_opt_run(vasp_command)
        else:
            print(f"Unsupported job type: {job}")
            sys.exit(-1)

        if not job_type.startswith("full_relax"):
            yield VaspJob(
                vasp_command,
                final=final,
                suffix=suffix,
                backup=backup,
                settings_override=settings,
                copy_magmom=copy_magmom,
                auto_npar=auto_npar,
            )


def do_run(args):
    """Do the run."""
    FORMAT = "%(asctime)s %(message)s"
    logging.basicConfig(format=FORMAT, level=logging.INFO, filename="run.log")
    logging.info(f"Handlers used are {args.handlers}")
    handlers = [load_class("custodian.vasp.handlers", n) for n in args.handlers]
    validators = [load_class("custodian.vasp.validators", n) for n in args.validators]

    c = Custodian(
        handlers,
        get_jobs(args),
        validators,
        max_errors=args.max_errors,
        scratch_dir=args.scratch,
        gzipped_output=args.gzip,
        checkpoint=True,
    )
    c.run()


def main():
    """Main method."""
    import argparse

    parser = argparse.ArgumentParser(
        description="run_vasp is a master script to perform various kinds of VASP runs.",
        epilog="Author: Shyue Ping Ong",
    )

    parser.add_argument(
        "-c",
        "--command",
        dest="command",
        nargs="?",
        default="pvasp",
        type=str,
        help="VASP command. Defaults to pvasp. If you are using mpirun, set this to something like 'mpirun pvasp'.",
    )

    parser.add_argument(
        "--no_auto_npar",
        action="store_true",
        help="Set to true to turn off auto_npar. Useful for certain machines "
        "and calculations where you want absolute control.",
    )

    parser.add_argument(
        "-z",
        "--gzip",
        dest="gzip",
        action="store_true",
        help="Add this option to gzip the final output. Do not gzip if you "
        "are going to perform an additional static run.",
    )

    parser.add_argument(
        "-s",
        "--scratch",
        dest="scratch",
        nargs="?",
        default=None,
        type=str,
        help="Scratch directory to perform run in. Specify the root scratch "
        "directory as the code will automatically create a temporary "
        "subdirectory to run the job.",
    )

    parser.add_argument(
        "-ks",
        "--kpoint-static",
        dest="static_kpoint",
        nargs="?",
        default=1,
        type=int,
        help="The multiplier to use for the KPOINTS of a static run (if "
        "any). For example, setting this to 2 means that if your "
        "original run was done using a k-point grid of 2x3x3, "
        "the static run will be done with a k-point grid of 4x6x6. This "
        "defaults to 1, i.e., static runs are performed with the same "
        "k-point grid as relaxation runs.",
    )

    parser.add_argument(
        "-me",
        "--max-errors",
        dest="max_errors",
        nargs="?",
        default=10,
        type=int,
        help="Maximum number of errors to allow before quitting",
    )

    parser.add_argument(
        "-hd",
        "--handlers",
        dest="handlers",
        nargs="+",
        default=[
            "VaspErrorHandler",
            "MeshSymmetryErrorHandler",
            "UnconvergedErrorHandler",
            "NonConvergingErrorHandler",
            "PotimErrorHandler",
        ],
        type=str,
        help="The ErrorHandlers to use specified as string class names, "
        "with optional arguments specified as a url-like string. For "
        "example, VaspErrorHandler?output_filename=myfile.out specifies a "
        "VaspErrorHandler with output_name set to myfile.out. Multiple "
        "arguments are joined by a comma. E.g., MyHandler?myfile=a,"
        "data=1. The arguments are deserialized using yaml.",
    )

    parser.add_argument(
        "-vd",
        "--validators",
        dest="validators",
        nargs="+",
        default=["VasprunXMLValidator"],
        type=str,
        help="The Validators to use specified as string class names, "
        "with optional arguments specified as a url-like string. For "
        "example, VaspErrorHandler?output_filename=myfile.out specifies a "
        "VaspErrorHandler with output_name set to myfile.out. Multiple "
        "arguments are joined by a comma. E.g., MyHandler?myfile=a,"
        "data=1. The arguments are deserialized using yaml.",
    )

    parser.add_argument(
        "jobs",
        metavar="jobs",
        type=str,
        nargs="+",
        default=["relax", "relax"],
        help="Jobs to execute. Only sequences of relax, "
        "quickrelax, static, rampU, full_relax, static_derived, "
        "nonscf_derived, optics_derived are "
        'supported at the moment. For example, "relax relax static" '
        "will run a double relaxation followed by a static "
        "run. By default, suffixes are given sequential numbering,"
        "but this can be overridden by adding a number to the job"
        "type, e.g. relax5 relax6 relax7",
    )

    args = parser.parse_args()
    do_run(args)
